
import os
import configparser
import logging
import json

from paramiko.client import SSHClient
from paramiko import AutoAddPolicy

from launchpadlib.launchpad import Launchpad
from launchpadlib.uris import lookup_service_root
from launchpadlib import credentials

from prometheus_client import Gauge


BUG_IMPORTANCE = [
   'Unknown',
   'Undecided',
   'Critical',
   'High',
   'Medium',
   'Low',
   'Wishlist',
]

BUG_STATUS = [
    'New',
    'Incomplete',
    'Triaged',
    'In Progress',
    'Confirmed',
]


def get_push_gateway(pkg, name):
    """Find push gateway in config file."""
    return get_config(pkg, name)['push-gateway']


def get_config(pkg, name):
    """
    Get configuration options for this script
    """
    config = configparser.SafeConfigParser()
    conffiles = [
        '/etc/{}.ini'.format(pkg),
        os.path.expanduser('~/.{}.ini'.format(pkg)),
        '{}.ini'.format(pkg),
    ]
    config.read(conffiles)
    return config[name]


class ShutUpAndTakeMyTokenAuthorizationEngine(credentials.RequestTokenAuthorizationEngine):

    """This stub class prevents launchpadlib from nulling out consumer_name
    in its demented campaign to force the use of desktop integration. """

    def __init__(self, service_root, application_name=None, consumer_name=None,
                 credential_save_failed=None, allow_access_levels=None):
        super(ShutUpAndTakeMyTokenAuthorizationEngine, self).__init__(
              service_root, application_name, consumer_name,
              credential_save_failed)


def launchpad_login(pkg):
    """Log into Launchpad API with stored credentials."""
    creds_dir = os.path.expanduser(os.path.join('~', '.' + pkg))
    if not os.path.exists(creds_dir):
        os.makedirs(creds_dir, 0o700)
    os.chmod(creds_dir, 0o700)
    api_endpoint = lookup_service_root('production')
    return Launchpad.login_with(
        application_name=pkg,
        credentials_file=os.path.join(creds_dir, 'launchpad.credentials'),
        service_root=api_endpoint,
        version='devel',
    )


def count_distro_bugs(distro, package):
    """Return the number of bugs for a given package."""
    source = distro.getSourcePackage(name=package)
    bugs = source.searchTasks()
    return len(bugs)


def gather_tagged_bugs(registry, projects, tags,
                       label, description):
    """Gather bugs tagged with any of tags for a list of projects."""
    gauge = Gauge(label, description,
                  ['project', 'bug_tag'],
                  registry=registry)
    for project in projects:
        for tag in tags:
            logging.info('Searching for bugs tagged'
                         ' with {} for {}'.format(tag, project))
            bugs = project.searchTasks(tags=[tag],
                                       status=BUG_STATUS)
            gauge.labels(project.name, tag).set(len(bugs))


def gather_project_bugs(registry, projects,
                        label, description):
    """Gather bugs by importance for a list of projects."""
    gauge = Gauge(label, description,
                  ['project', 'importance'],
                  registry=registry)
    for project in projects:
        logging.info('Searching for bugs for {}'.format(project))
        for importance in BUG_IMPORTANCE:
            gauge.labels(
                project.name,
                importance).set(len(
                    project.searchTasks(importance=importance,
                                        status=BUG_STATUS
                )))

def gather_project_bugs_by_importance(registry, projects,
                                      label, description):
    """Gather bugs by importance for all projects."""
    gauge = Gauge(label, description,
                  ['importance'],
                  registry=registry)
    for importance in BUG_IMPORTANCE:
        count = 0
        for project in projects:
            logging.info('Searching for bugs for {}'.format(project))
            count += len(
                project.searchTasks(importance=importance,
                                    status=BUG_STATUS)
            ) 
        gauge.labels(importance).set(count)


def gather_new_bugs(registry, projects,
                    label, description):
    """Count new bugs associated with a list of projects."""
    bug_count = 0
    for project in projects:
        bug_count += len(project.searchTasks(status='New'))

    Gauge(label, description, None,
          registry=registry).set(bug_count)


def gather_bug_reporters(registry, projects, members,
                         label, description):
    """Gather bugs by reporter type for a list of projects."""
    gauge = Gauge(label, description,
                  ['project', 'reporter_type'],
                  registry=registry)
    for project in projects:
        bugs = project.searchTasks()
        canonical_bugs = []
        for member in members:
            if member.account_status != 'Active':
                continue
            bugs = project.searchTasks(bug_reporter=member)
            canonical_bugs.extend(bugs)

        bug_count = len(bugs)
        canonical_count = len(canonical_bugs)
        non_canonical_count = bug_count - canonical_count

        gauge.labels(project.name, 'internal').set(canonical_count)
        gauge.labels(project.name, 'external').set(non_canonical_count)

# NOTE(jamespage): Minimal Gerrit SSH client lifted from pygerrit
def escape_string(string):
    """ Escape a string for use in Gerrit commands.
    :arg str string: The string to escape.
    :returns: The string with necessary escapes and surrounding double quotes
        so that it can be passed to any of the Gerrit commands that require
        double-quoted strings.
    """

    result = string
    result = result.replace('\\', '\\\\')
    result = result.replace('"', '\\"')
    return '"' + result + '"'


class GerritSSHClient(SSHClient):

    def __init__(self, hostname, username=None, port=None,
                 keepalive=None, key_filename=None, proxy=None):
        """ Initialise and connect to SSH. """
        super(GerritSSHClient, self).__init__()
        self.remote_version = None
        self.hostname = hostname
        self.username = username
        self.key_filename = key_filename
        self.port = port
        self.connected = False
        self.proxy = proxy
        self.keepalive = keepalive

    def _do_connect(self):
        """ Connect to the remote. """
        self.load_system_host_keys()
        if self.username is None or self.port is None:
            self._configure()

        self.set_missing_host_key_policy(AutoAddPolicy())
        self.connect(hostname=self.hostname,
                     port=self.port,
                     username=self.username,
                     key_filename=self.key_filename,
                     sock=self.proxy)

    def _connect(self):
        if not self.connected:
            self._do_connect()

    def query(self, term):
        self._connect()

        command = ["query", "--current-patch-set", "--all-approvals",
                   "--format JSON", "--commit-message"]
        command.append(escape_string(term))
        command = "gerrit " + " ".join(command)

        try:
            command.encode('ascii')
        except UnicodeEncodeError:
            command = command.encode('utf-8')

        _, stdout, _ = self.exec_command(command,
                                         bufsize=1,
                                         timeout=None,
                                         get_pty=False)

        results = []
        for line in stdout.read().splitlines():
            data = json.loads(line.decode('UTF-8'))
            if 'project' in data:
                results.append(data)
        return results


def gather_review_stats(registry, config, name,
                        gerrit_args, queries):
    gerrit_args['username'] = config['gerrit-username']
    client = GerritSSHClient(**gerrit_args)

    for qtype in queries:
        count = len(client.query(queries[qtype]))
        print('{}: {}'.format(qtype, count))
        Gauge('openstack_{}_{}_reviews'.format(name, qtype),
              'Reviews ({})'.format(qtype), None,
              registry=registry).set(count)
