#!/usr/bin/env python3

"""
Copyright (c) 2016 Canonical, Ltd.
Authors: Paul Gear, James Page

Hooks for openstack-kpi-scripts charm.
"""

import base64
import glob
import os
import re

from charmhelpers.core import (
    host,
    hookenv,
    unitdata,
)

from charms.reactive import (
    main,
    remove_state,
    set_state,
)

from charms.reactive.decorators import (
    hook,
    when,
    when_all,
    when_not,
)

from charmhelpers.fetch import apt_install

from charmhelpers.core.templating import render

from charmhelpers.core.host import write_file


CHARM_NAME = 'openstack-kpi-scripts'


def status(status, msg):
    hookenv.log('%s: %s' % (status, msg))
    hookenv.status_set(status, msg)


def active(msg):
    status('active', msg)


def blocked(msg):
    status('blocked', msg)


def maint(msg):
    status('maintenance', msg)


def write_config_file():
    """
    Create /etc/openstack-kpi-scripts.ini.
    """
    cfg_file = '{}.ini'.format(CHARM_NAME)
    maint('rendering config %s' % (cfg_file,))
    script_dir = '/srv/{}/parts'.format(CHARM_NAME)
    scripts = [x for x in os.listdir(
        script_dir) if re.match(r'^[-_A-Za-z]+$', x)]
    render(
        source=cfg_file,
        target='/etc/' + cfg_file,
        perms=0o755,
        context={
            'push_gateway': hookenv.config('prometheus-push-gateway'),
            'scripts': scripts,
            'github_token': hookenv.config('github-token'),
            'config': hookenv.config('') or {},
            'gerrit_username': hookenv.config('gerrit-username'),
        },
    )
    return hookenv.config('prometheus-push-gateway')


def write_cron_job():
    """
    Create cron job
    """
    dst = '/etc/cron.d/{}'.format(CHARM_NAME)
    cron_job = 'cron-job'
    maint('installing %s to %s' % (cron_job, dst))
    kv = unitdata.kv()
    render(
        source=cron_job,
        target=dst,
        perms=0o755,
        context={
            'script_dir': '/srv/{}/parts'.format(CHARM_NAME),
            'script_name': CHARM_NAME,
            'user': kv.get('run-as'),
        },
    )


def write_launchpad_credentials():
    """
    Save launchpad credentials in the configured file.
    Data comes in as base64 string, so we must convert to bytes,
    decode, then convert back to string before writing to the file.
    """
    kv = unitdata.kv()
    creds_file = kv.get('launchpad-credentials-file')
    maint('saving Launchpad credentials to %s' % (creds_file,))
    dst_dir = os.path.dirname(creds_file)
    user = kv.get('run-as')
    host.mkdir(dst_dir, owner=user, perms=0o700)
    creds_blob = kv.get('launchpad-credentials')
    creds_data = base64.b64decode(creds_blob.encode())
    host.write_file(creds_file, creds_data, owner=user)


def install_ssh_private_key():
    write_file('/home/ubuntu/.ssh/id_rsa_gerrit',
               str.encode(hookenv.config('gerrit-ssh-private-key')),
               owner='ubuntu', group='ubuntu', perms=0o600)


@when_all(
    '{}.configured'.format(CHARM_NAME),
)
def write_config():
    blocked('Unable to configure charm - please see log')
    write_launchpad_credentials()
    install_ssh_private_key()
    push_gateway = write_config_file()
    write_cron_job()
    active('Configured push gateway %s' % (push_gateway,))


@hook('config-changed')
def config_changed():
    remove_state('{}.configured'.format(CHARM_NAME))
    maint('checking configuration')

    kv = unitdata.kv()
    config_items = (
        'launchpad-credentials',
        'launchpad-credentials-file',
        'run-as',
        'github-token',
        'prometheus-push-gateway',
        'gerrit-ssh-private-key',
        'gerrit-username',
    )
    for c in config_items:
        item = hookenv.config(c)
        if not item:
            blocked('%s must be set' % (c,))
            remove_state('{}.configured'.format(CHARM_NAME))
            return
        else:
            kv.set(c, item)
    set_state('{}.configured'.format(CHARM_NAME))


@hook(
    'install',
    'upgrade-charm',
)
def install_files():
    # this part lifted from haproxy charm hooks.py
    src = os.path.join(os.environ["CHARM_DIR"], "files/thirdparty/")
    dst = '/srv/{}/parts/'.format(CHARM_NAME)
    maint('Copying scripts from %s to %s' % (src, dst))
    host.mkdir(dst, perms=0o755)
    for fname in glob.glob(os.path.join(src, "*")):
        host.rsync(fname, os.path.join(dst, os.path.basename(fname)))

    # Template files may have changed in an upgrade, so we need to rewrite
    # them
    config_changed()

    # Package prerequisites for openstack-kpi-scripts/thirdparty/*
    apt_install([
        'python3-prometheus-client',
        'python3-github',
        'python3-cssselect',
        'python3-yaml',
        'python3-launchpadlib',
        'python3-urllib3',
        'python3-bs4',
        'python3-paramiko',
        ])
